"""
Cryptoauthlib Device Configuration
"""
# (c) 2015-2018 Microchip Technology Inc. and its subsidiaries.
#
# Subject to your compliance with these terms, you may use Microchip software
# and any derivatives exclusively with Microchip products. It is your
# responsibility to comply with third party license terms applicable to your
# use of third party software (including open source software) that may
# accompany Microchip software.
#
# THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
# EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
# WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
# PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT,
# SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE
# OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF
# MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE
# FORESEEABLE. TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL
# LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED
# THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR
# THIS SOFTWARE.


from ctypes import Structure, c_uint16, c_uint8
from .library import AtcaStructure

# Because this module directly mirrors the C api the following is an exception to the python coding standard
# pylint: disable-msg=too-few-public-methods


class AesEnable(Structure):
    """AES Enable (608) Field Definition """
    _fields_ = [('Enable', c_uint8, 1),
                ('Reserved', c_uint8, 6)]
    _pack_ = 1


class I2cEnable(Structure):
    """I2C Enable Field Definition """
    _fields_ = [('Enable', c_uint8, 1),
                ('Reserved', c_uint8, 6)]
    _pack_ = 1


class CountMatch(Structure):
    """CountMatch (608) Field Definition """
    _fields_ = [('Enable', c_uint8, 1),
                ('Reserved', c_uint8, 3),
                ('CountMatchKey', c_uint8, 4)]
    _pack_ = 1


class ChipMode508(Structure):
    """ChipMode for 508 Field Definition """
    _fields_ = [('UserExtraAdd', c_uint8, 1),
                ('TTLenable', c_uint8, 1),
                ('WatchdogDuration', c_uint8, 1)]
    _pack_ = 1


class ChipMode608(Structure):
    """ChipMode for 608 Field Definition """
    _fields_ = [('UserExtraAdd', c_uint8, 1),
                ('TTLenable', c_uint8, 1),
                ('WatchdogDuration', c_uint8, 1),
                ('ClockDivider', c_uint8, 5)]
    _pack_ = 1


class Counter204(Structure):
    """Counter Definition for SHA204 """
    _fields_ = [('UseFlag', c_uint8),
                ('UpdateCount', c_uint8)]
    _pack_ = 1


class SlotConfig(Structure):
    """Slot Configuration Field Definition """
    _fields_ = [('ReadKey', c_uint16, 4),
                ('NoMac', c_uint16, 1),
                ('LimitedUse', c_uint16, 1),
                ('EncryptRead', c_uint16, 1),
                ('IsSecret', c_uint16, 1),
                ('WriteKey', c_uint16, 4),
                ('WriteConfig', c_uint16, 4)]
    _pack_ = 1


class UseLock(Structure):
    """UseLock Field Definition """
    _fields_ = [('UseLockEnable', c_uint8, 4),
                ('UseLockKey', c_uint8, 4)]
    _pack_ = 1


class VolatileKeyPermission(Structure):
    """VolatileKeyPermission Field Definition """
    _fields_ = [('VolatileKeyPermitSlot', c_uint8, 4),
                ('Reserved', c_uint8, 3),
                ('VolatileKeyPermitEnable', c_uint8, 1)]
    _pack_ = 1


class SecureBoot(Structure):
    """SecureBoot Field Definition """
    _fields_ = [('SecureBootMode', c_uint16, 2),
                ('Reserved0', c_uint16, 1),
                ('SecureBootPersistentEnable', c_uint16, 1),
                ('SecureBootRandNonce', c_uint16, 1),
                ('Reserved1', c_uint16, 3),
                ('SecureBootSigDig', c_uint16, 4),
                ('SecureBootPubKey', c_uint16, 4)]
    _pack_ = 1


class ChipOptions(Structure):
    """ChipOptions Field Definition """
    _fields_ = [('PowerOnSelfTest', c_uint16, 1),
                ('IoProtectionKeyEnable', c_uint16, 1),
                ('KdfAesEnable', c_uint16, 1),
                ('AutoClearFirstFail', c_uint16, 1),
                ('Reserved', c_uint16, 4),
                ('EcdhProtectionBits', c_uint16, 2),
                ('KdfProtectionBits', c_uint16, 2),
                ('IoProtectionKey', c_uint16, 4)]
    _pack_ = 1


class X509Format(Structure):
    """X509Format Field Definition """
    _fields_ = [('PublicPosition', c_uint8, 4),
                ('TemplateLength', c_uint8, 4)]
    _pack_ = 1


class KeyConfig(Structure):
    """KeyConfig Field Definition """
    _fields_ = [('Private', c_uint16, 1),
                ('PubInfo', c_uint16, 1),
                ('KeyType', c_uint16, 3),
                ('Lockable', c_uint16, 1),
                ('ReqRandom', c_uint16, 1),
                ('ReqAuth', c_uint16, 1),
                ('AuthKey', c_uint16, 4),
                ('PersistentDisable', c_uint16, 1),
                ('RFU', c_uint16, 1),
                ('X509id', c_uint16, 2)]
    _pack_ = 1


class Atsha204aConfig(AtcaStructure):
    """ATSHA204A Config Zone Definition """
    _fields_ = [('SN03', c_uint8*4),
                ('RevNum', c_uint8*4),
                ('SN48', c_uint8*5),
                ('Reserved13', c_uint8),
                ('I2C_Enable', I2cEnable),
                ('Reserved15', c_uint8),
                ('I2C_Address', c_uint8),
                ('CheckMacConfig', c_uint8),
                ('OTPmode', c_uint8),
                ('SelectorMode', c_uint8),
                ('SlotConfig', SlotConfig*16),
                ('Counter', Counter204*8),
                ('LastKeyUse', c_uint8*16),
                ('UserExtra', c_uint8),
                ('Selector', c_uint8),
                ('LockValue', c_uint8),
                ('LockConfig', c_uint8)]
    _pack_ = 1


class Atecc508aConfig(AtcaStructure):
    """ATECC508A Config Zone Definition """
    _fields_ = [('SN03', c_uint8*4),
                ('RevNum', c_uint8*4),
                ('SN48', c_uint8*5),
                ('Reserved13', c_uint8),
                ('I2C_Enable', I2cEnable),
                ('Reserved15', c_uint8),
                ('I2C_Address', c_uint8),
                ('Reserved17', c_uint8),
                ('OTPmode', c_uint8),
                ('ChipMode', ChipMode508),
                ('SlotConfig', SlotConfig*16),
                ('Counter0', c_uint8*8),
                ('Counter1', c_uint8*8),
                ('LastKeyUse', c_uint8*16),
                ('UserExtra', c_uint8),
                ('Selector', c_uint8),
                ('LockValue', c_uint8),
                ('LockConfig', c_uint8),
                ('SlotLocked', c_uint16),
                ('RFU', c_uint16),
                ('X509format', X509Format*4),
                ('KeyConfig', KeyConfig*16)]
    _pack_ = 1


class Atecc608Config(AtcaStructure):
    """ATECC608 Config Zone Definition """
    _fields_ = [('SN03', c_uint8*4),
                ('RevNum', c_uint8*4),
                ('SN48', c_uint8*5),
                ('AES_Enable', AesEnable),
                ('I2C_Enable', I2cEnable),
                ('Reserved15', c_uint8),
                ('I2C_Address', c_uint8),
                ('Reserved17', c_uint8),
                ('CountMatch', CountMatch),
                ('ChipMode', ChipMode608),
                ('SlotConfig', SlotConfig*16),
                ('Counter0', c_uint8*8),
                ('Counter1', c_uint8*8),
                ('UseLock', UseLock),
                ('VolatileKeyPermission', VolatileKeyPermission),
                ('SecureBoot', SecureBoot),
                ('KdfIvLoc', c_uint8),
                ('KdfIvStr', c_uint8*2),
                ('Reserved68', c_uint8*9),
                ('UserExtra', c_uint8),
                ('UserExtraAdd', c_uint8),
                ('LockValue', c_uint8),
                ('LockConfig', c_uint8),
                ('SlotLocked', c_uint16),
                ('ChipOptions', ChipOptions),
                ('X509format', X509Format*4),
                ('KeyConfig', KeyConfig*16)]
    _pack_ = 1


__all__ = ['Atsha204aConfig', 'Atecc508aConfig', 'Atecc608Config']



"""
TNG: classes and functions for interacting with TNG devices
"""
# (c) 2015-2019 Microchip Technology Inc. and its subsidiaries.
#
# Subject to your compliance with these terms, you may use Microchip software
# and any derivatives exclusively with Microchip products. It is your
# responsibility to comply with third party license terms applicable to your
# use of third party software (including open source software) that may
# accompany Microchip software.
#
# THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
# EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
# WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
# PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT,
# SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE
# OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF
# MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE
# FORESEEABLE. TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL
# LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED
# THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR
# THIS SOFTWARE.

from ctypes import c_int, c_size_t, byref, create_string_buffer
from .library import get_cryptoauthlib, AtcaReference
from .atcaenum import AtcaEnum
from .status import Status

# Because this module directly mirrors the C api the following is an exception to the python coding standard
# pylint: disable-msg=invalid-name


def tng_get_device_pubkey(public_key):
    """
    Uses GenKey command to calculate the public key from the primary
    device public key.

    Args:
        public_key  Public key will be returned here. Format will be
                    the X and Y integers in big-endian format.
                    64 bytes for P256 curve. Expects bytearray.

    Returns:
        ATCA_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(public_key, bytearray):
        status = Status.ATCA_BAD_PARAM
    else:
        c_public_key = create_string_buffer(64)
        status = get_cryptoauthlib().tng_get_device_pubkey(byref(c_public_key))
        public_key[0:] = bytes(c_public_key.raw)
    return status


def tng_atcacert_max_device_cert_size(max_cert_size):
    """
    Return the maximum possible certificate size in bytes for a TNG
    device certificate. Certificate can be variable size, so this
    gives an appropriate buffer size when reading the certificate.

    Args:
        max_cert_size  Maximum certificate size will be returned here
                       in bytes. Expects AtcaReference.
    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(max_cert_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_max_cert_size = c_size_t(0)
        status = get_cryptoauthlib().tng_atcacert_max_device_cert_size(byref(c_max_cert_size))
        max_cert_size.value = c_max_cert_size.value
    return status


def tng_atcacert_read_device_cert(cert, cert_size, signer_cert=None):
    """
    Reads the device certificate for a TNG device.

    Args:
        cert         Buffer to received the certificate (DER format).
                     Expects bytearray.
        cert_size    As input, the size of the cert buffer in bytes.
                     As output, the size of the certificate returned
                     in cert in bytes. Expects AtcaReference.
        signer_cert  If supplied, the signer public key is used from
                     this certificate. If set to None, the signer
                     public key is read from the device.
                     Expects bytes or None.
    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(cert, bytearray) or not isinstance(cert_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_cert = create_string_buffer(cert_size.value)
        c_cert_size = c_size_t(cert_size.value)
        if signer_cert is not None:
            signer_cert = bytes(signer_cert)
        status = get_cryptoauthlib().tng_atcacert_read_device_cert(byref(c_cert), byref(c_cert_size), signer_cert)
        cert[:] = bytes(c_cert.raw)[0:c_cert_size.value]
        cert_size.value = c_cert_size.value
    return status


def tng_atcacert_device_public_key(public_key, cert=None):
    """
    Reads the device public key.

    Args:
        public_key   Public key will be returned here. Format will be
                     the X and Y integers in big-endian format.
                     64 bytes for P256 curve. Expects bytearray.
        cert         If supplied, the device public key is used from
                     this certificate. If set to None, the device
                     public key is read from the device. Expects bytes or None.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(public_key, bytearray):
        status = Status.ATCA_BAD_PARAM
    else:
        c_public_key = create_string_buffer(64)
        if cert is not None:
            cert = bytes(cert)
        status = get_cryptoauthlib().tng_atcacert_device_public_key(byref(c_public_key), cert)
        public_key[:] = bytes(c_public_key.raw)
    return status


def tng_atcacert_max_signer_cert_size(max_cert_size):
    """
    Return the maximum possible certificate size in bytes for a TNG
    signer certificate. Certificate can be variable size, so this
    gives an appropriate buffer size when reading the certificate.

    Args:
        max_cert_size  Maximum certificate size will be returned here
                       in bytes. Expects AtcaReference.
    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(max_cert_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_max_cert_size = c_size_t(0)
        status = get_cryptoauthlib().tng_atcacert_max_signer_cert_size(byref(c_max_cert_size))
        max_cert_size.value = c_max_cert_size.value
    return status


def tng_atcacert_read_signer_cert(cert, cert_size):
    """
    Reads the signer certificate for a TNG device.

    Args:
        cert         Buffer to received the certificate (DER format).
                     Expects bytearray.
        cert_size    As input, the size of the cert buffer in bytes.
                     As output, the size of the certificate returned
                     in cert in bytes. Expects AtcaReference.
    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(cert, bytearray) or not isinstance(cert_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_cert = create_string_buffer(cert_size.value)
        c_cert_size = c_size_t(cert_size.value)
        status = get_cryptoauthlib().tng_atcacert_read_signer_cert(byref(c_cert), byref(c_cert_size))
        cert[:] = bytes(c_cert.raw)[0:c_cert_size.value]
        cert_size.value = c_cert_size.value
    return status


def tng_atcacert_signer_public_key(public_key, cert=None):
    """
    Reads the signer public key.

    Args:
        public_key   Public key will be returned here. Format will be
                     the X and Y integers in big-endian format.
                     64 bytes for P256 curve. Expects bytearray.
        cert         If supplied, the signer public key is used from
                     this certificate. If set to None, the signer
                     public key is read from the device. Expects bytes or None.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(public_key, bytearray):
        status = Status.ATCA_BAD_PARAM
    else:
        c_public_key = create_string_buffer(64)
        if cert is not None:
            cert = bytes(cert)
        status = get_cryptoauthlib().tng_atcacert_signer_public_key(byref(c_public_key), cert)
        public_key[:] = bytes(c_public_key.raw)
    return status


def tng_atcacert_root_cert_size(cert_size):
    """
    Get the size of the TNG root cert.

    Args:
        cert_size  Certificate size will be returned here in bytes.
                   Expects AtcaReference.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(cert_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_cert_size = c_size_t(0)
        status = get_cryptoauthlib().tng_atcacert_root_cert_size(byref(c_cert_size))
        cert_size.value = c_cert_size.value
    return status


def tng_atcacert_root_cert(cert, cert_size):
    """
    Get the TNG root cert.

    Args:
        cert       Buffer to received the certificate (DER format).
                   Expects bytearray.
        cert_size  As input, the size of the cert buffer in bytes.
                   As output, the size of the certificate returned
                   in cert in bytes. Expects AtcaReference.
    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(cert, bytearray) or not isinstance(cert_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_cert = create_string_buffer(cert_size.value)
        c_cert_size = c_size_t(cert_size.value)
        status = get_cryptoauthlib().tng_atcacert_root_cert(byref(c_cert), byref(c_cert_size))
        cert[:] = bytes(c_cert.raw)[0:c_cert_size.value]
        cert_size.value = c_cert_size.value
    return status


def tng_atcacert_root_public_key(public_key):
    """
    Gets the root public key.

    Args:
        public_key  Public key will be returned here. Format will be
                    the X and Y integers in big-endian format.
                    64 bytes for P256 curve. Expects bytearray.
    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(public_key, bytearray):
        status = Status.ATCA_BAD_PARAM
    else:
        c_public_key = create_string_buffer(64)
        status = get_cryptoauthlib().tng_atcacert_root_public_key(byref(c_public_key))
        public_key[:] = bytes(c_public_key.raw)
    return status



"""
Enum Extension for improved comparisons
"""
from enum import Enum

# This is because @DynamicClassAttribue isn't exactly @property and pylint doesn't really understand it as the same
# pylint: disable-msg=comparison-with-callable


class AtcaEnum(Enum):
    """
    Overload of standard python enum for some additional convenience features. Assumes closer alignment to C style enums
    where the value is always an integer
    """
    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, str):
            answer = (self.name == other)
        else:
            answer = (self.value == int(other))
        return answer

    def __ne__(self, other):
        if isinstance(other, str):
            answer = (self.name != other)
        else:
            answer = (self.value != int(other))
        return answer

    def __int__(self):
        return int(self.value)


# Make module import * safe - keep at the end of the file
__all__ = ['AtcaEnum']



"""
ATCACERT: classes and functions for interacting with compressed certificates
"""
# (c) 2015-2018 Microchip Technology Inc. and its subsidiaries.
#
# Subject to your compliance with these terms, you may use Microchip software
# and any derivatives exclusively with Microchip products. It is your
# responsibility to comply with third party license terms applicable to your
# use of third party software (including open source software) that may
# accompany Microchip software.
#
# THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
# EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
# WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
# PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT,
# SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE
# OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF
# MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE
# FORESEEABLE. TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL
# LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED
# THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR
# THIS SOFTWARE.

import binascii
from datetime import datetime
from ctypes import Structure, c_int, c_uint8, c_uint16, c_char, POINTER, create_string_buffer, c_uint32, byref, c_size_t
from .library import get_cryptoauthlib, get_ctype_by_name, AtcaReference, AtcaStructure
from .atcaenum import AtcaEnum
from .status import Status

# Because this module directly mirrors the C api the following is an exception to the python coding standard
# pylint: disable-msg=invalid-name, too-few-public-methods


class atcacert_cert_type_t(AtcaEnum):
    """
    Types of certificates
    """
    CERTTYPE_X509 = 0       # Standard X509 certificate
    CERTTYPE_CUSTOM = 1     # Custom format


class atcacert_cert_sn_src_t(AtcaEnum):
    """
    Sources for the certificate serial number
    """
    # Cert serial is stored on the device.
    SNSRC_STORED = 0x0
    # Cert serial is stored on the device with the first byte being the DER size (X509 certs only).
    SNSRC_STORED_DYNAMIC = 0x7
    # Cert serial number is 0x40(MSB) + 9-byte device serial number. Only applies to device certificates.
    SNSRC_DEVICE_SN = 0x8
    # Cert serial number is 0x40(MSB) + 2-byte signer ID. Only applies to signer certificates.
    SNSRC_SIGNER_ID = 0x9
    # Cert serial number is the SHA256(Subject public key + Encoded dates), with uppermost 2 bits set to 01.
    SNSRC_PUB_KEY_HASH = 0xA
    # Cert serial number is the SHA256(Device SN + Encoded dates), with uppermost 2 bits set to 01.
    # Only applies to device certificates.
    SNSRC_DEVICE_SN_HASH = 0xB
    # Depreciated, don't use. Cert serial number is the SHA256(Subject public key + Encoded dates),
    # with MSBit set to 0 to ensure it's positive.
    SNSRC_PUB_KEY_HASH_POS = 0xC
    # Depreciated, don't use. Cert serial number is the SHA256(Device SN + Encoded dates),
    # with MSBit set to 0 to ensure it's positive. Only applies to device certificates.
    SNSRC_DEVICE_SN_HASH_POS = 0xD
    # Depreciated, don't use. Cert serial number is the SHA256(Subject public key + Encoded dates).
    SNSRC_PUB_KEY_HASH_RAW = 0xE
    # Depreciated, don't use. Cert serial number is the SHA256(Device SN + Encoded dates).
    # Only applies to device certificates.
    SNSRC_DEVICE_SN_HASH_RAW = 0xF


class atcacert_device_zone_t(AtcaEnum):
    """
    ATECC device zones. The values match the Zone Encodings as specified in the datasheet
    """
    DEVZONE_CONFIG = 0x00   # Configuration zone.
    DEVZONE_OTP = 0x01      # One Time Programmable zone.
    DEVZONE_DATA = 0x02     # Data zone (slots).
    DEVZONE_NONE = 0x07     # Special value used to indicate there is no device location.


class atcacert_transform_t(AtcaEnum):
    """
    Transforms for converting the device data.
    """
    TF_NONE = 0x00              # No transform, data is used byte for byte
    TF_REVERSE = 0x01           # Reverse the bytes (e.g. change endianess)
    TF_BIN2HEX_UC = 0x02        # Convert raw binary into ASCII hex, uppercase
    TF_BIN2HEX_LC = 0x03        # Convert raw binary into ASCII hex, lowercase
    TF_HEX2BIN_UC = 0x04        # Convert ASCII hex, uppercase to binary
    TF_HEX2BIN_LC = 0x05        # Convert ASCII hex, lowercase to binary
    TF_BIN2HEX_SPACE_UC = 0x06  # Convert raw binary into ASCII hex, uppercase space between bytes
    TF_BIN2HEX_SPACE_LC = 0x07  # Convert raw binary into ASCII hex, lowercase space between bytes
    TF_HEX2BIN_SPACE_UC = 0x08  # Convert ASCII hex, uppercase with spaces between bytes to binary
    TF_HEX2BIN_SPACE_LC = 0x09  # Convert ASCII hex, lowercase with spaces between bytes to binary

class atcacert_date_format_t(AtcaEnum):
    """
    Support Date formats by the atcacert
    """
    # ISO8601 full date YYYY-MM-DDThh:mm:ssZ
    DATEFMT_ISO8601_SEP = 0
    # RFC 5280 (X.509) 4.1.2.5.1 UTCTime format YYMMDDhhmmssZ
    DATEFMT_RFC5280_UTC = 1
    # POSIX (aka UNIX) date format. Seconds since Jan 1, 1970. 32 bit unsigned integer, big endian.
    DATEFMT_POSIX_UINT32_BE = 2
    # POSIX (aka UNIX) date format. Seconds since Jan 1, 1970. 32 bit unsigned integer, little endian.
    DATEFMT_POSIX_UINT32_LE = 3
    # RFC 5280 (X.509) 4.1.2.5.2 GeneralizedTime format YYYYMMDDhhmmssZ
    DATEFMT_RFC5280_GEN = 4


class atcacert_std_cert_element_t(AtcaEnum):
    """
    Standard dynamic certificate elements
    """
    STDCERT_PUBLIC_KEY = 0
    STDCERT_SIGNATURE = 1
    STDCERT_ISSUE_DATE = 2
    STDCERT_EXPIRE_DATE = 3
    STDCERT_SIGNER_ID = 4
    STDCERT_CERT_SN = 5
    STDCERT_AUTH_KEY_ID = 6
    STDCERT_SUBJ_KEY_ID = 7


class CertStatus(AtcaEnum):
    """
    Status codes returned from atcacert commands and their meanings. From atcacert.h
    """
    ATCACERT_E_SUCCESS = 0               # Operation completed successfully.
    ATCACERT_E_ERROR = 1                 # General error.
    ATCACERT_E_BAD_PARAMS = 2            # Invalid/bad parameter passed to function.
    ATCACERT_E_BUFFER_TOO_SMALL = 3      # Supplied buffer for output is too small to hold the result.
    ATCACERT_E_DECODING_ERROR = 4        # Data being decoded/parsed has an invalid format.
    ATCACERT_E_INVALID_DATE = 5          # Date is invalid.
    ATCACERT_E_UNIMPLEMENTED = 6         # Function is unimplemented for the current configuration.
    ATCACERT_E_UNEXPECTED_ELEM_SIZE = 7  # A certificate element size was not what was expected.
    ATCACERT_E_ELEM_MISSING = 8          # The certificate element isn't defined for the certificate definition.
    ATCACERT_E_ELEM_OUT_OF_BOUNDS = 9    # Certificate element is out of bounds for the given certificate.
    ATCACERT_E_BAD_CERT = 10             # Certificate structure is bad in some way.
    ATCACERT_E_WRONG_CERT_DEF = 11
    ATCACERT_E_VERIFY_FAILED = 12        # Certificate or challenge/response verification failed.


def _atcacert_convert_bytes(kwargs, name, pointer):
    """
    Internal Helper Function: Convert python 'bytes' into memory pointer for ctypes structure
    :param kwargs: kwargs dictionary
    :param name: _field_ name that will be converted
    :param pointer: Conversion Class (resulting type - pointer of type x)
    :return:
    """
    k = kwargs.get(name)
    if k is not None:
        k = k.replace(' ', '').strip()
        byte_string = binascii.unhexlify(k)
        kwargs[name] = pointer((c_uint8*len(byte_string))(*list(byte_string)))


def _atcacert_convert_enum(kwargs, name, enum):
    """
    Internal Helper Function: Convert python enum into ctypes integer
    :param kwargs: kwargs dictionary
    :param name: _field_ name that will be converted
    :param enum: Conversion Class (resulting type)
    :return:
    """
    k = kwargs.get(name)
    if k is not None and k is not int:
        if isinstance(k, enum):
            kwargs[name] = int(k)
        else:
            kwargs[name] = int(getattr(enum, k))


class atcacert_device_loc_t(AtcaStructure):
    """
    CTypes mirror of atcacert_device_loc_t from atcacert_def.h
    """
    _fields_ = [
        ('zone', get_ctype_by_name('atcacert_device_zone_t')),  # Zone in the device.
        ('slot', c_uint8),  # Slot within the data zone. Only applies if zone is DEVZONE_DATA.
        ('is_genkey', c_uint8),  # If true, use GenKey command to get the contents instead of Read.
        ('offset', c_uint16),  # Byte offset in the zone.
        ('count', c_uint16)  # Byte count.
    ]
    _pack_ = 1

    def __init__(self, *args, **kwargs):
        if kwargs is not None:
            _atcacert_convert_enum(kwargs, 'zone', atcacert_device_zone_t)

        super(atcacert_device_loc_t, self).__init__(*args, **kwargs)


class atcacert_cert_loc_t(AtcaStructure):
    """
    CTypes mirror of atcacert_cert_loc_t from atcacert_def.h
    """
    _fields_ = [('offset', c_uint16), ('count', c_uint16)]
    _pack_ = 1


class atcacert_cert_element_t(AtcaStructure):
    """
    CTypes mirror of atcacert_cert_element_t from atcacert_def.h
    """
    _fields_ = [
        ('id', c_char * 25),  # ID identifying this element.
        ('device_loc', atcacert_device_loc_t),  # Location in the device for the element.
        ('cert_loc', atcacert_cert_loc_t),  # Location in the certificate template for the element.
        ('transforms', get_ctype_by_name('atcacert_transform_t') * 2)  # Transforms for converting the device data.

    ]
    _pack_ = 1


class atcacert_def_t(AtcaStructure):
    """
    CTypes mirror of atcacert_def_t from atcacert_def.h
    """
    def __init__(self, *args, **kwargs):
        if kwargs is not None:
            _atcacert_convert_enum(kwargs, 'type', atcacert_cert_type_t)
            _atcacert_convert_enum(kwargs, 'sn_source', atcacert_cert_sn_src_t)
            _atcacert_convert_enum(kwargs, 'issue_date_format', atcacert_date_format_t)
            _atcacert_convert_enum(kwargs, 'expire_date_format', atcacert_date_format_t)

            _atcacert_convert_bytes(kwargs, 'cert_template', POINTER(c_uint8))

        super(atcacert_def_t, self).__init__(*args, **kwargs)

# Need to define fields outside the class due to ca_cert_def, which is a pointer
# to the same class.
atcacert_def_t._fields_ = [  # pylint: disable=protected-access
    # Certificate type.
    ('type', get_ctype_by_name('atcacert_cert_type_t')),
    # ID for the this certificate definition (4-bit value).
    ('template_id', c_uint8),
    # ID for the certificate chain this definition is a part of (4-bit value).
    ('chain_id', c_uint8),
    # If this is a device certificate template, this is the device slot for the device private key.
    ('private_key_slot', c_uint8),
    # Where the certificate serial number comes from (4-bit value).
    ('sn_source', get_ctype_by_name('atcacert_cert_sn_src_t')),
    # Only applies when sn_source is SNSRC_STORED or SNSRC_STORED_DYNAMIC. Describes where to get the
    # certificate serial number on the device.
    ('cert_sn_dev_loc', atcacert_device_loc_t),
    # Format of the issue date in the certificate.
    ('issue_date_format', get_ctype_by_name('atcacert_date_format_t')),
    # format of the expire date in the certificate.
    ('expire_date_format', get_ctype_by_name('atcacert_date_format_t')),
    # Location in the certificate for the TBS (to be signed) portion.
    ('tbs_cert_loc', atcacert_cert_loc_t),
    # Number of years the certificate is valid for (5-bit value). 0 means no expiration.
    ('expire_years', c_uint8),
    # Where on the device the public key can be found.
    ('public_key_dev_loc', atcacert_device_loc_t),
    # Where on the device the compressed cert can be found.
    ('comp_cert_dev_loc', atcacert_device_loc_t),
    # Where in the certificate template the standard cert elements are inserted.
    ('std_cert_elements', atcacert_cert_loc_t * 8),
    # Additional certificate elements outside of the standard certificate contents.
    ('cert_elements', POINTER(atcacert_cert_element_t)),
    # Number of additional certificate elements in cert_elements.
    ('cert_elements_count', c_uint8),
    # Pointer to the actual certificate template data.
    ('cert_template', POINTER(c_uint8)),
    # Size of the certificate template in cert_template in bytes.
    ('cert_template_size', c_uint16),
    # Certificate definition of the CA certificate
    ('ca_cert_def', POINTER(atcacert_def_t))
]

class atcacert_tm_utc_t(Structure):
    """
    CTypes mirror of atcacert_tm_utc_t from atcacert_date.h which mimics the posix time structure
    """
    _fields_ = [
        ('tm_sec', c_int),   # 0 to 59
        ('tm_min', c_int),   # 0 to 59
        ('tm_hour', c_int),  # 0 to 23
        ('tm_mday', c_int),  # 1 to 31
        ('tm_mon', c_int),   # 0 to 11
        ('tm_year', c_int),  # years since 1900
    ]

    def __init__(self, *args, **kwargs):
        if not args and not kwargs:
            t = datetime.utcnow()
            args = (t.second, t.minute, t.hour, t.day, t.month - 1, t.year)

        super(atcacert_tm_utc_t, self).__init__(*args, **kwargs)

        if self.tm_sec not in range(60):
            raise ValueError('tm_sec out of range: {}'.format(self.tm_sec))
        if self.tm_min not in range(60):
            raise ValueError('tm_min out of range: {}'.format(self.tm_min))
        if self.tm_hour not in range(24):
            raise ValueError('tm_hour out of range: {}'.format(self.tm_hour))
        if self.tm_mday not in range(1, 31):
            raise ValueError('tm_mday out of range: {}'.format(self.tm_mday))
        if self.tm_mon not in range(12):
            raise ValueError('tm_mon out of range: {}'.format(self.tm_mon))
        # pylint: disable=no-member
        # pylint thinks tm_year isn't a member for some reason
        if self.tm_year not in range(200):
            if self.tm_year >= 1900:
                self.tm_year -= 1900
            else:
                raise ValueError('tm_year out of range: {}'.format(self.tm_year))


#===============================================================================
# Client side cert i/o methods. These declarations deal with the client-side, the node being authenticated,
# of the authentication process. It is assumed the client has an ECC CryptoAuthentication device
# (e.g. ATECC508A) and the certificates are stored on that device.


def atcacert_max_cert_size(cert_def, max_cert_size):
    """
    Return the maximum possible certificate size in bytes for a given
    cert def. Certificate can be variable size, so this gives an
    appropriate buffer size when reading the certificates.

    Args:
        cert_def       Certificate definition to find a max size for.
                       Expects atcacert_def_t.
        max_cert_size  Maximum certificate size will be returned here in bytes.
                       Expects AtcaReference.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(max_cert_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_max_cert_size = c_size_t(0)
        status = get_cryptoauthlib().atcacert_max_cert_size(byref(cert_def), byref(c_max_cert_size))
        max_cert_size.value = c_max_cert_size.value
    return status


def atcacert_get_response(device_private_key_slot, challenge, response):
    """
    Calculates the response to a challenge sent from the host.
    The challenge-response protocol is an ECDSA Sign and Verify. This performs the ECDSA Sign on the
    challenge and returns the signature as the response.

    Args:
        device_private_key_slot         Slot number for the device's private key. This must be the
                                        same slot used to generate the public key included in the
                                        device's certificate.
        challenge                       Challenge to generate the response for. Must be 32 bytes.
        response                        Response will be returned in this buffer. 64 bytes.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(response, bytearray):
        status = Status.ATCA_BAD_PARAM
    else:
        c_response = create_string_buffer(64)
        status = get_cryptoauthlib().atcacert_get_response(device_private_key_slot, bytes(challenge), byref(c_response))
        response[0:] = bytes(c_response.raw)
    return status


def atcacert_read_cert(cert_def, ca_public_key, cert, cert_size):
    """
    Reads the certificate specified by the certificate definition from the
    ATECC508A device.
    This process involves reading the dynamic cert data from the device and combining it
    with the template found in the certificate definition.

    Args:
        cert_def                Certificate definition describing where to find the dynamic
                                certificate information on the device and how to incorporate it
                                into the template. Expects atcacert_def_t.
        ca_public_key           The ECC P256 public key of the certificate authority that signed
                                this certificate. Formatted as the 32 byte X and Y integers
                                concatenated together (64 bytes total). Set to NULL if the
                                authority key id is not needed, set properly in the cert_def
                                template, or stored on the device as specifed in the
                                cert_def cert_elements.
        cert                    Buffer to received the certificate. Expects bytearray.
        cert_size               As input, the size of the cert buffer in bytes.
                                As output, the size of the certificate returned in cert in bytes.
                                Expects AtcaReference.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(cert, bytearray) or not isinstance(cert_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_cert_size = c_uint32(cert_size.value)
        c_cert = create_string_buffer(cert_size.value)
        status = get_cryptoauthlib().atcacert_read_cert(byref(cert_def), bytes(ca_public_key), byref(c_cert),
                                                        byref(c_cert_size))
        cert[:] = bytes(c_cert.raw)[0:c_cert_size.value]
        cert_size.value = c_cert_size.value
    return status


def atcacert_write_cert(cert_def, cert, cert_size):
    """
    Take a full certificate and write it to the ATECC508A device according to the
    certificate definition.

    Args:
        cert_def                Certificate definition describing where the dynamic certificate
                                information is and how to store it on the device.
                                Expects atcacert_def_t.
        cert                    Full certificate to be stored.
        cert_size               Size of the full certificate in bytes.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    status = get_cryptoauthlib().atcacert_write_cert(byref(cert_def), bytes(cert), cert_size)
    return status


def atcacert_create_csr(csr_def, csr, csr_size):
    """
    Creates a CSR specified by the CSR definition from the ATECC508A device.
    This process involves reading the dynamic CSR data from the device and combining it
    with the template found in the CSR definition, then signing it. Return the CSR int der format

    Args:
        csr_def                 CSR definition describing where to find the dynamic CSR information
                                on the device and how to incorporate it into the template.
                                Expects atcacert_def_t.
        csr                     Buffer to receive the CSR. Expects bytearray.
        csr_size                As input, the size of the CSR buffer in bytes.
                                As output, the size of the CSR as PEM returned in cert in bytes.
                                Expects AtcaReference.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(csr, bytearray) or not isinstance(csr_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_csr_size = c_uint32(csr_size.value)
        c_csr = create_string_buffer(csr_size.value)
        status = get_cryptoauthlib().atcacert_create_csr(byref(csr_def), byref(c_csr), byref(c_csr_size))
        csr[:] = bytes(c_csr.raw)[0:c_csr_size.value]
        csr_size.value = c_csr_size.value
    return status


def atcacert_create_csr_pem(csr_def, csr, csr_size):
    """
    Creates a CSR specified by the CSR definition from the ATECC508A device.
    This process involves reading the dynamic CSR data from the device and combining it
    with the template found in the CSR definition, then signing it. Return the CSR int der format

    Args:
        csr_def                 CSR definition describing where to find the dynamic CSR information
                                on the device and how to incorporate it into the template.
                                Expects atcacert_def_t.
        csr                     Buffer to receive the CSR. Expects bytearray.
        csr_size                As input, the size of the CSR buffer in bytes.
                                As output, the size of the CSR as PEM returned in cert in bytes.
                                Expects AtcaReference.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(csr, bytearray) or not isinstance(csr_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_csr_size = c_uint32(csr_size.value)
        c_csr = create_string_buffer(csr_size.value)
        status = get_cryptoauthlib().atcacert_create_csr(byref(csr_def), byref(c_csr), byref(c_csr_size))
        csr[:] = bytes(c_csr.raw)[0:c_csr_size.value]
        csr_size.value = c_csr_size.value
    return status


# atcacert_date
# Date handling with regard to certificates.


def atcacert_date_enc(date_format, timestamp, formatted_date, formatted_date_size):
    """
    Format a timestamp according to the format type.

    Args:
        date_format             Format to use.
        timestamp               Timestamp to format. Expects atcacert_tm_utc_t.
        formatted_date          Formatted date will be returned in this buffer.
                                Expects bytearray.
        formatted_date_size     As input, the size of the formatted_date buffer.
                                As output, the size of the returned formatted_date.
                                Expects AtcaReference.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    if not isinstance(formatted_date, bytearray) or not isinstance(formatted_date_size, AtcaReference):
        status = Status.ATCA_BAD_PARAM
    else:
        c_formatted_date = create_string_buffer(formatted_date_size.value)
        c_formatted_date_size = c_uint32(formatted_date_size.value)
        status = get_cryptoauthlib().atcacert_date_enc(int(date_format), byref(timestamp), byref(c_formatted_date),
                                                       byref(c_formatted_date_size))
        formatted_date[:] = bytes(c_formatted_date.raw)[0:c_formatted_date_size.value]
        formatted_date_size.value = c_formatted_date_size.value
    return status


def atcacert_date_dec(date_format, formatted_date, formatted_date_size, timestamp):
    """
    Parse a formatted timestamp according to the specified format.

    Args:
        date_format             Format to parse the formatted date as.
        formatted_date          Formatted date to be parsed.
        formatted_date_size     Size of the formatted date in bytes.
        timestamp               Parsed timestamp is returned here. Expects atcacert_tm_utc_t.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    status = get_cryptoauthlib().atcacert_date_dec(int(date_format), bytes(formatted_date), formatted_date_size,
                                                   byref(timestamp))
    return status


def atcacert_date_enc_compcert(issue_date, expire_years, enc_dates):
    """
    Encode the issue and expire dates in the format used by the compressed certificate.

    Args:
        issue_date              Issue date to encode. Note that minutes and seconds will be ignored.
                                Expects atcacert_tm_utc_t.
        expire_years            Expire date is expressed as a number of years past the issue date.
                                0 should be used if there is no expire date.
        enc_dates               Encoded dates for use in the compressed certificate is returned here.
                                3 bytes. Expects bytearray.

    Returns:
        ATCACERT_E_SUCCESS on success
    """
    if not isinstance(enc_dates, bytearray):
        status = Status.ATCA_BAD_PARAM
    else:
        c_enc_dates = create_string_buffer(3)
        status = get_cryptoauthlib().atcacert_date_enc_compcert(byref(issue_date), expire_years, byref(c_enc_dates))
        enc_dates[0:] = bytes(c_enc_dates.raw)
    return status


def atcacert_date_dec_compcert(enc_dates, expire_date_format, issue_date, expire_date):
    """
    Decode the issue and expire dates from the format used by the compressed certificate.

    Args:
        enc_dates               Encoded date from the compressed certificate. 3 bytes.
        expire_date_format      Expire date format. Only used to determine max date when no
                                expiration date is specified by the encoded date.
        issue_date              Decoded issue date is returned here. Expects atcacert_tm_utc_t.
        expire_date             Decoded expire date is returned here. If there is no
                                expiration date, the expire date will be set to a maximum
                                value for the given expire_date_format. Expects atcacert_tm_utc_t.
    Returns:
        ATCACERT_E_SUCCESS on success
    """
    status = get_cryptoauthlib().atcacert_date_dec_compcert(bytes(enc_dates), int(expire_date_format),
                                                            byref(issue_date), byref(expire_date))
    return status


def atcacert_date_get_max_date(date_format, timestamp):
    """
    Return the maximum date available for the given format.

    Args:
        format                  Format to get the max date for.
        timestamp               Max date is returned here. Expects atcacert_tm_utc_t.

    Returns:
        ATCACERT_E_SUCCESS on success, otherwise an error code.
    """
    status = get_cryptoauthlib().atcacert_date_get_max_date(int(date_format), byref(timestamp))
    return status


# Make module import * safe - keep at the end of the file
__all__ = [x for x in dir() if x.startswith(__name__.split('.')[-1])] + ['CertStatus']



from setuptools import setup, Distribution
from setuptools.command.install import install
from setuptools.command.build_ext import build_ext
from setuptools.extension import Extension

import sys
import subprocess
import os
import glob
import shutil
import re
from ctypes import cdll

_NAME = 'cryptoauthlib'
_DESCRIPTION = 'Python Wrapper Library for Microchip Security Products'
_AUTHOR = 'Microchip Technology Inc'
_AUTHOR_EMAIL = 'support@microchip.com'
_LICENSE = 'Other'
_VERSION = ""
_URL = 'https://github.com/MicrochipTech/cryptoauthlib'

_DOWNLOAD_URL = '%s/archive/%s.tar.gz' % (_URL, _VERSION)
_CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: Other/Proprietary License',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Operating System :: OS Independent',
]

_PROJECT_URLS = {
    'Documentation': '%s/wiki/python' % _URL,
    'Source': _URL,
    'Tracker': '%s/issues' % _URL,
}

# Include the compiled library in the resulting distribution
_PACKAGE_DATA = {'cryptoauth_signatures': ['cryptoauth.json']}
if sys.platform == 'win32':
    _PACKAGE_DATA['libcryptoauth'] = ['cryptoauth.dll']
#elif sys.platform is 'darwin':
else:
    _PACKAGE_DATA['libcryptoauth'] = ['libcryptoauth.so']

# See if this is being built from an sdist structure
if os.path.exists('lib') and os.path.exists('third_party'):
    _sdist_build = True
else:
    _sdist_build = False

#python 2 compatibility
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

# See if the library is already installed
try:
    lib = cdll.LoadLibrary('libcryptoauth.so')
    # Test to ensure it has the required features to support the
    # python wrapper. It may change later to a version check
    assert 0 != lib.ATCAIfaceCfg_size
    _EXTENSIONS = None
except:
    _EXTENSIONS = [Extension('cryptoauthlib', sources=[])]

# Try to load the version
try:
    _VERSION = open('VERSION', 'r').read().strip()
except FileNotFoundError:
    with open('lib/atca_version.h', 'r') as f:
        m = re.search(r'ATCA_LIBRARY_VERSION_DATE\s+\"([0-9]+)\"', f.read(), re.M)
        _VERSION = m.groups()[0]


def copy_udev_rules(target):
    if _sdist_build:
        rules = 'lib/hal/90-cryptohid.rules'
    else:
        rules = '../lib/hal/90-cryptohid.rules'

    if not os.path.exists(target):
        raise FileNotFoundError

    if not os.path.exists(target + os.path.sep + os.path.basename(rules)):
        shutil.copy(rules, target)


def install_udev_rules():
    if sys.platform.startswith('linux'):
        try:
            copy_udev_rules('/etc/udev/rules.d')
        except PermissionError:
            print('Unable to write udev rules. Rerun install as sudo or install rules manually')
        except:
            print('Unable to install udev rules. See readme to manually install')


def load_readme():
    with open('README.md', 'r') as f:
        read_me = f.read()

    return read_me


class CryptoAuthCommandBuildExt(build_ext):
    def build_extension(self, ext):
        # Suppress cmake output
        devnull = open(os.devnull, 'r+b')
        nousb = bool(os.environ.get('CRYPTOAUTHLIB_NOUSB', False))
        cmake_gen = os.environ.get('CMAKE_GEN', None)

        # Check if CMAKE is installed
        try:
            subprocess.check_call(['cmake', '--version'], stdin=devnull, stdout=devnull, stderr=devnull, shell=False)
        except OSError as e:
            print("CMAKE must be installed on the system for this module to build the required extension e.g. 'apt-get install cmake' or 'yum install cmake'")
            raise e

        extdir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name)) + os.path.sep + _NAME)
        setupdir = os.path.dirname(os.path.abspath(__file__)) + os.path.sep

        cmakelist_path = os.path.abspath(setupdir + 'lib' if _sdist_build else '../lib')

        if not sys.platform.startswith('linux'):
            cfg = 'Debug' if self.debug else 'Release'
            build_args = ['--config', cfg]
        else:
            build_args = []

        cmake_args = ['-DATCA_HAL_CUSTOM=ON', '-DATCA_HAL_KIT_UART=ON',
                      '-DATCA_ATSHA206A_SUPPORT=ON',
                      '-DATCA_TNGTLS_SUPPORT=ON', '-DATCA_TFLEX_SUPPORT=ON',
                      '-DATCA_TNGLORA_SUPPORT=ON', '-DATCA_TNG_LEGACY_SUPPORT=ON',
                      '-DATCA_USE_ATCAB_FUNCTIONS=ON']
                      
        if os.path.exists('../lib/talib' if not _sdist_build else 'lib/talib'):
            cmake_args += ['-DATCA_TA100_SUPPORT=ON']
            if sys.platform.startswith('linux'):
                cmake_args += ['-DATCA_OPENSSL=ON']
            else:
                cmake_args += ['-DATCA_TA100_AES_AUTH_SUPPORT=OFF']

        if not nousb:
            cmake_args += ['-DATCA_HAL_KIT_HID=ON']

        if 'win32' == sys.platform:
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_%s=' % cfg.upper() + extdir,
                          '-DCMAKE_RUNTIME_OUTPUT_DIRECTORY_%s=' % cfg.upper() + extdir]
            if cmake_gen is not None:
                cmake_args += ['-G', cmake_gen]
            elif sys.maxsize > 2**32:
                    cmake_args += ['-A', 'x64']
            else:
                    cmake_args += ['-A', 'win32']
        else:
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir]

        if sys.platform.startswith('linux'):
            cmake_args += ['-DATCA_HAL_I2C=ON', '-DATCA_HAL_SPI=ON']

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        # Configure the library
        try:
            subprocess.check_output(['cmake', cmakelist_path] + cmake_args, cwd=os.path.abspath(self.build_temp), stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            msg = e.output.decode('utf-8')
            if 'usb' in msg:
                msg += '\n\n   USB libraries or headers were not located. If USB support is\n' \
                       '   not required it can be disabled by setting the environment\n' \
                       '   variable CRYPTOAUTHLIB_NOUSB to true before trying to install\n' \
                       '   this package: \n\n' \
                       '       $ export CRYPTOAUTHLIB_NOUSB=True\n\n' \
                       '   Run setup.py clean before trying install again or use the pip \n' \
                       '   option --no-cache-dir\n'
            raise RuntimeError(msg)

        # Build the library
        try:
            subprocess.check_output(['cmake', '--build', '.'] + build_args, cwd=os.path.abspath(self.build_temp), stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            if sys.version_info[0] <= 2:
                raise RuntimeError(e.output)  # Python 2 doesn't handle unicode exceptions
            else:
                raise RuntimeError(e.output.decode('utf-8'))


class CryptoAuthCommandInstall(install):
    def run(self):
        self.do_egg_install()
        install_udev_rules()


class BinaryDistribution(Distribution):
    def has_ext_modules(self):
        return (_EXTENSIONS is not None)

# Setuptools has some weird behavior when the install command class is extended
# but only affects bdist_* invocations which only applies to macos and windows
# and the extension is only required for linux
_COMMANDS = { 'build_ext': CryptoAuthCommandBuildExt }
#if sys.platform.startswith('linux'):
#    _COMMANDS['install'] = CryptoAuthCommandInstall

if __name__ == '__main__':
    setup(
        name=_NAME,
        packages=[_NAME],
        version=_VERSION,
        description=_DESCRIPTION,
        long_description=load_readme(),
        long_description_content_type='text/markdown',
        url=_URL,
        author=_AUTHOR,
        author_email=_AUTHOR_EMAIL,
        download_url=_DOWNLOAD_URL,
        keywords='Microchip ATECC508A ATECC608 ECDSA ECDH',
        project_urls=_PROJECT_URLS,
        license=_LICENSE,
        classifiers=_CLASSIFIERS,
        package_data=_PACKAGE_DATA,
        include_package_data=True,
        distclass=BinaryDistribution,
        cmdclass=_COMMANDS,
        setup_requires=['setuptools>=38.6.0', 'wheel'],
        install_requires=['enum34;python_version<"3.4"'],
        ext_modules=_EXTENSIONS,
        python_requires='>=3.4.*',
        zip_safe=False
    )



"""
Cryptoauthlib Exceptions
"""
# (c) 2015-2018 Microchip Technology Inc. and its subsidiaries.
#
# Subject to your compliance with these terms, you may use Microchip software
# and any derivatives exclusively with Microchip products. It is your
# responsibility to comply with third party license terms applicable to your
# use of third party software (including open source software) that may
# accompany Microchip software.
#
# THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
# EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
# WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
# PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT,
# SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE
# OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF
# MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE
# FORESEEABLE. TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL
# LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED
# THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR
# THIS SOFTWARE.


class CryptoError(Exception):
    """Standard CryptoAuthLib Exceptions"""


class LibraryLoadError(CryptoError):
    """CryptpAuthLib failed to Load"""


class ConfigZoneLockedError(CryptoError):
    """Config Zone Locked"""


class DataZoneLockedError(CryptoError):
    """Configuration Enabled"""


class WakeFailedError(CryptoError):
    """Device Wake failed"""


class CheckmacVerifyFailedError(CryptoError):
    """response status byte indicates CheckMac failure (status byte = 0x01)"""


class ParseError(CryptoError):
    """response status byte indicates parsing error (status byte = 0x03)"""


class CrcError(CryptoError):
    """response status byte indicates CRC error (status byte = 0xFF)"""


class StatusUnknownError(CryptoError):
    """Response status byte is unknown"""


class EccFaultError(CryptoError):
    """response status byte is ECC fault (status byte = 0x05)"""


class FunctionError(CryptoError):
    """Function could not execute due to incorrect condition / state."""


class GenericError(CryptoError):
    """unspecified error"""


class BadArgumentError(CryptoError):
    """bad argument (out of range, null pointer, etc.)"""


class InvalidIdentifierError(CryptoError):
    """invalid device id, id not set"""


class InvalidSizeError(CryptoError):
    """Count value is out of range or greater than buffer size."""


class BadCrcError(CryptoError):
    """incorrect CRC received"""


class ReceiveError(CryptoError):
    """Timed out while waiting for response. Number of bytes received is > 0."""


class NoResponseError(CryptoError):
    """error while the Command layer is polling for a command response."""


class ResyncWithWakeupError(CryptoError):
    """Re-synchronization succeeded, but only after generating a Wake-up"""


class ParityError(CryptoError):
    """for protocols needing parity"""


class TransmissionTimeoutError(CryptoError):
    """for Microchip PHY protocol, timeout on transmission waiting for master"""


class ReceiveTimeoutError(CryptoError):
    """for Microchip PHY protocol, timeout on receipt waiting for master"""


class CommunicationError(CryptoError):
    """Communication with device failed. Same as in hardware dependent modules."""


class TimeOutError(CryptoError):
    """Timed out while waiting for response. Number of bytes received is 0."""


class BadOpcodeError(CryptoError):
    """Opcode is not supported by the device"""


class ExecutionError(CryptoError):
    """chip was in a state where it could not execute the command, response
    status byte indicates command execution error (status byte = 0x0F)"""


class UnimplementedError(CryptoError):
    """Function or some element of it hasn't been implemented yet"""


class AssertionFailure(CryptoError):
    """Code failed run-time consistency check"""


class TransmissionError(CryptoError):
    """Failed to write"""


class ZoneNotLockedError(CryptoError):
    """required zone was not locked"""


class NoDevicesFoundError(CryptoError):
    """For protocols that support device discovery (kit protocol), no devices were found"""


class HealthTestError(CryptoError):
    """Random number generator health test error"""


class LibraryMemoryError(CryptoError):
    """CryptoAuthLib was unable to allocate memory"""


class NoUseFlagError(CryptoError):
    """Indication that no dk pk flag is available to perform"""


class LibraryNotInitialized(CryptoError):
    """Indication that library or context was not initialized prior to an API call"""

class UnsupportedInterface(CryptoError):
    """"The selected interface is not supported by the library"""



"""
Cryptoauthlib Library Management
"""
# (c) 2015-2018 Microchip Technology Inc. and its subsidiaries.
#
# Subject to your compliance with these terms, you may use Microchip software
# and any derivatives exclusively with Microchip products. It is your
# responsibility to comply with third party license terms applicable to your
# use of third party software (including open source software) that may
# accompany Microchip software.
#
# THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
# EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
# WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
# PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT,
# SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE
# OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF
# MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE
# FORESEEABLE. TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL
# LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED
# THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR
# THIS SOFTWARE.

import os.path
import sys
from ctypes import *
from ctypes.util import find_library
from .exceptions import LibraryLoadError
from .atcaenum import AtcaEnum

# Maps common name to the specific name used internally
ATCA_NAMES = {'i2c': 'i2c', 'hid': 'kithid', 'sha': 'sha204', 'ecc': 'eccx08'}

# Global cdll instance of the loaded compiled library
_CRYPTO_LIB = None

# List of basic ctypes by size
_CTYPES_BY_SIZE = {1: c_uint8, 2: c_uint16, 4:c_uint32}

class AtcaReference:
    """
    A simple wrapper to pass an immutable type to a function for return
    """
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __gt__(self, other):
        return self.value > other

    def __ge__(self, other):
        return self.value >= other

    def __int__(self):
        return int(self.value)

    def __str__(self):
        return str(self.value)


def _force_local_library():
    """
    In some environments loading seems to fail under all circumstances unless
    brute forcing it.
    """
    curr_path = os.path.dirname(__file__)

    if sys.platform.startswith('win'):
        return os.path.join(curr_path, "cryptoauth.dll")
    elif sys.platform.startswith('darwin'):
        return os.path.join(curr_path, "libcryptoauth.dylib")
    else:
        return os.path.join(curr_path, "libcryptoauth.so")


def load_cryptoauthlib(lib=None):
    """
    Load CryptoAauthLib into Python environment
    raise LibraryLoadError if cryptoauthlib library can't be loaded
    """
    global _CRYPTO_LIB      # pylint: disable=global-statement
    if lib is not None:
        _CRYPTO_LIB = lib
    else:
        try:
            library_file = find_library('cryptoauth')
            if library_file is None:
                library_file = _force_local_library()
            _CRYPTO_LIB = cdll.LoadLibrary(library_file)
        except:
            try:
                _CRYPTO_LIB = cdll.LoadLibrary(_force_local_library())
            except:
                raise LibraryLoadError('Unable to find cryptoauthlib. You may need to reinstall')



def get_cryptoauthlib():
    """
    This is a helper function for the other python files in this module to use the loaded library
    """
    global _CRYPTO_LIB      # pylint: disable=global-statement
    return _CRYPTO_LIB


def get_device_name(revision):
    """
    Returns the device name based on the info byte array values returned by atcab_info
    """
    devices = {0x10: 'ATECC108A',
               0x50: 'ATECC508A',
               0x60: 'ATECC608',
               0x20: 'ECC204',
               0x00: 'ATSHA204A',
               0x02: 'ATSHA204A',
               0x40: 'ATSHA206A'}
    device_name = devices.get(revision[2], 'UNKNOWN')
    return device_name


def get_device_type_id(name):
    """
    Returns the ATCADeviceType value based on the device name
    """
    devices = {'ATSHA204A': 0,
               'ATECC108A': 1,
               'ATECC508A': 2,
               'ATECC608A': 3,
               'ATECC608B': 3,
               'ATECC608': 3,
               'ATSAH206A': 4,
               'ECC204': 5,
               'UNKNOWN': 0x20}
    return devices.get(name.upper())


def get_size_by_name(name):
    """
    Get the size of an object in the library using the name_size api from atca_utils_sizes.c
    """
    global _CRYPTO_LIB      # pylint: disable=global-statement
    return getattr(_CRYPTO_LIB, '{}_size'.format(name), lambda: 4)()


def get_ctype_by_name(name):
    """
    For known (atca_utils_sizes.c) types that are custom to the library retrieve the size
    """
    return _CTYPES_BY_SIZE.get(get_size_by_name(name))


def get_ctype_structure_instance(structure, value):
    """
    Internal Helper Function:  Convert a value into the correct ctypes structure for a given field
    :param value: Value to convert
    :param structure: Conversion Class (resulting type)
    :return:
    """
    # pylint: disable-msg=invalid-name
    if isinstance(value, dict):
        r = structure(**value)
    elif isinstance(value, int):
        r = structure.from_buffer_copy(c_uint(value))
    elif isinstance(value, AtcaEnum):
        r = structure.from_buffer_copy(c_uint(int(value)))
    elif not isinstance(value, structure):
        r = structure(value)
    else:
        r = value
    return r


def get_ctype_array_instance(array, value):
    """
    Internal Helper Function: Convert python list into ctype array
    :param value: Value to convert
    :param array: Conversion Class (resulting type)
    :return:
    """
    # pylint: disable-msg=invalid-name, protected-access
    t = array._type_
    if t is c_char:
        # Strings are special
        if isinstance(value, str):
            a = value.encode('ascii')
        else:
            a = bytes(value)
    else:
        a = array(*[get_ctype_structure_instance(t, e) for e in value])
    return a


class AtcaUnion(Union):
    """ An extended ctypes structure to accept complex inputs """
    # pylint: disable-msg=invalid-name, too-few-public-methods
    def __init__(self, *args, **kwargs):
        if kwargs is not None:
            for f in self._fields_:
                if f[0] in kwargs:
                    if isinstance(f[1](), Union):
                        kwargs[f[0]] = get_ctype_structure_instance(f[1], kwargs[f[0]])
                    elif isinstance(f[1](), Structure):
                        kwargs[f[0]] = get_ctype_structure_instance(f[1], kwargs[f[0]])
                    elif isinstance(f[1](), Array):
                        kwargs[f[0]] = get_ctype_array_instance(f[1], kwargs[f[0]])

        super(AtcaUnion, self).__init__(*args, **kwargs)


class AtcaStructure(Structure):
    """ An extended ctypes structure to accept complex inputs """
    # pylint: disable-msg=invalid-name, too-few-public-methods
    def __init__(self, *args, **kwargs):
        if kwargs is not None:
            for f in self._fields_:
                if f[0] in kwargs:
                    if isinstance(f[1](), Union):
                        kwargs[f[0]] = get_ctype_structure_instance(f[1], kwargs[f[0]])
                    elif isinstance(f[1](), Structure):
                        kwargs[f[0]] = get_ctype_structure_instance(f[1], kwargs[f[0]])
                    elif isinstance(f[1](), Array):
                        kwargs[f[0]] = get_ctype_array_instance(f[1], kwargs[f[0]])

        super(AtcaStructure, self).__init__(*args, **kwargs)

    def update_from_buffer(self, buffer):
        if len(buffer) < sizeof(self):
            raise ValueError
        memmove(addressof(self), buffer, sizeof(self))


def ctypes_to_bytes(obj):
    """
    Convert a ctypes structure/array into bytes. This is for python2 compatibility
    """
    buf = create_string_buffer(sizeof(obj))
    memmove(buf, addressof(obj), sizeof(obj))
    return buf.raw


def create_byte_buffer(init_or_size):
    if isinstance(init_or_size, int):
        buf = (c_uint8*init_or_size)()
    else:
        buf = (c_uint8*len(init_or_size))(*list(init_or_size))
    return buf


__all__ = ['ATCA_NAMES', 'AtcaReference', 'load_cryptoauthlib', 'get_device_name', 'get_device_type_id',
           'create_byte_buffer']



"""
Status codes and status to exception conversions.
"""
# (c) 2015-2018 Microchip Technology Inc. and its subsidiaries.
#
# Subject to your compliance with these terms, you may use Microchip software
# and any derivatives exclusively with Microchip products. It is your
# responsibility to comply with third party license terms applicable to your
# use of third party software (including open source software) that may
# accompany Microchip software.
#
# THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
# EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
# WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
# PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT,
# SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE
# OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF
# MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE
# FORESEEABLE. TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL
# LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED
# THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR
# THIS SOFTWARE.

# Because all modules are setup to be import * safe the following is an exception to the python coding standard
# pylint: disable-msg=wildcard-import, unused-wildcard-import

from .atcaenum import AtcaEnum
from .exceptions import *

class Status(AtcaEnum):
    """
    Status codes returned from cryptoauthlib commands and their meanings. See atca_status.h
    """
    # Success
    ATCA_SUCCESS = 0x00
    # Config Zone Locked
    ATCA_CONFIG_ZONE_LOCKED = 0x01
    # Configuration Enabled
    ATCA_DATA_ZONE_LOCKED = 0x02
    # Response status byte indicates CheckMac failure (status byte = 0x01)
    ATCA_WAKE_FAILED = 0xD0
    # Response status byte indicates CheckMac failure (status byte = 0x01)
    ATCA_CHECKMAC_VERIFY_FAILED = 0xD1
    # Response status byte indicates parsing error (status byte = 0x03)
    ATCA_PARSE_ERROR = 0xD2
    # response status byte indicates CRC error (status byte = 0xFF)
    ATCA_STATUS_CRC = 0xD4
    # response status byte is unknown
    ATCA_STATUS_UNKNOWN = 0xD5
    # response status byte is ECC fault (status byte = 0x05)
    ATCA_STATUS_ECC = 0xD6
    # Function could not execute due to incorrect condition / state.
    ATCA_FUNC_FAIL = 0xE0
    # unspecified error
    ATCA_GEN_FAIL = 0xE1
    # bad argument (out of range, null pointer, etc.)
    ATCA_BAD_PARAM = 0xE2
    # invalid device id, id not set
    ATCA_INVALID_ID = 0xE3
    # Count value is out of range or greater than buffer size.
    ATCA_INVALID_SIZE = 0xE4
    # incorrect CRC received
    ATCA_BAD_CRC = 0xE5
    # Timed out while waiting for response. Number of bytes received is > 0.
    ATCA_RX_FAIL = 0xE6
    # Not an error while the Command layer is polling for a command response.
    ATCA_RX_NO_RESPONSE = 0xE7
    # Re-synchronization succeeded, but only after generating a Wake-up
    ATCA_RESYNC_WITH_WAKEUP = 0xE8
    # For protocols needing parity
    ATCA_PARITY_ERROR = 0xE9
    # For Microchip Kit PHY protocol, timeout on transmission waiting for master
    ATCA_TX_TIMEOUT = 0xEA
    # For Microchip Kit PHY protocol, timeout on receipt waiting for master
    ATCA_RX_TIMEOUT = 0xEB
    # Communication with device failed. Same as in hardware dependent modules.
    ATCA_COMM_FAIL = 0xF0
    # Timed out while waiting for response. Number of bytes received is 0.
    ATCA_TIMEOUT = 0xF1
    # opcode is not supported by the device
    ATCA_BAD_OPCODE = 0xF2
    # Received proper wake token
    ATCA_WAKE_SUCCESS = 0xF3
    # Chip was in a state where it could not execute the command, response status byte indicates command
    # execution error (status byte = 0x0F)
    ATCA_EXECUTION_ERROR = 0xF4
    # Function or some element of it hasn't been implemented yet
    ATCA_UNIMPLEMENTED = 0xF5
    # Code failed run-time consistency check
    ATCA_ASSERT_FAILURE = 0xF6
    # Failed to write
    ATCA_TX_FAIL = 0xF7
    # required zone was not locked
    ATCA_NOT_LOCKED = 0xF8
    # For protocols that support device discovery (kit protocol), no devices were found
    ATCA_NO_DEVICES = 0xF9
    # Random number generator health test error
    ATCA_HEALTH_TEST_ERROR = 0xFA
    # Couldn't allocate required memory
    ATCA_ALLOC_FAILURE = 0xFB
    # All dk pk flags are consumed so no use flag is available
    ATCA_USE_FLAGS_CONSUMED = 0xFC
    # The library has not been initialized so the command could not be executed
    ATCA_NOT_INITIALIZED = 0xFD


STATUS_EXCEPTION_MAP = {
    int(Status.ATCA_SUCCESS): None,
    int(Status.ATCA_CONFIG_ZONE_LOCKED): ConfigZoneLockedError,
    int(Status.ATCA_DATA_ZONE_LOCKED): DataZoneLockedError,
    int(Status.ATCA_WAKE_FAILED): WakeFailedError,
    int(Status.ATCA_CHECKMAC_VERIFY_FAILED): CheckmacVerifyFailedError,
    int(Status.ATCA_PARSE_ERROR): ParseError,
    int(Status.ATCA_STATUS_CRC): CrcError,
    int(Status.ATCA_STATUS_UNKNOWN): StatusUnknownError,
    int(Status.ATCA_STATUS_ECC): EccFaultError,
    int(Status.ATCA_FUNC_FAIL): FunctionError,
    int(Status.ATCA_GEN_FAIL): GenericError,
    int(Status.ATCA_BAD_PARAM): BadArgumentError,
    int(Status.ATCA_INVALID_ID): InvalidIdentifierError,
    int(Status.ATCA_INVALID_SIZE): InvalidSizeError,
    int(Status.ATCA_BAD_CRC): BadCrcError,
    int(Status.ATCA_RX_FAIL): ReceiveError,
    int(Status.ATCA_RX_NO_RESPONSE): NoResponseError,
    int(Status.ATCA_RESYNC_WITH_WAKEUP): ResyncWithWakeupError,
    int(Status.ATCA_PARITY_ERROR): ParityError,
    int(Status.ATCA_TX_TIMEOUT): TransmissionTimeoutError,
    int(Status.ATCA_RX_TIMEOUT): ReceiveTimeoutError,
    int(Status.ATCA_COMM_FAIL): CommunicationError,
    int(Status.ATCA_TIMEOUT): TimeOutError,
    int(Status.ATCA_BAD_OPCODE): BadOpcodeError,
    int(Status.ATCA_WAKE_SUCCESS): None,
    int(Status.ATCA_EXECUTION_ERROR): ExecutionError,
    int(Status.ATCA_UNIMPLEMENTED): UnimplementedError,
    int(Status.ATCA_ASSERT_FAILURE): AssertionFailure,
    int(Status.ATCA_TX_FAIL): TransmissionError,
    int(Status.ATCA_NOT_LOCKED): ZoneNotLockedError,
    int(Status.ATCA_NO_DEVICES): NoDevicesFoundError,
    int(Status.ATCA_HEALTH_TEST_ERROR): HealthTestError,
    int(Status.ATCA_ALLOC_FAILURE): LibraryMemoryError,
    int(Status.ATCA_USE_FLAGS_CONSUMED): NoUseFlagError,
    int(Status.ATCA_NOT_INITIALIZED): LibraryNotInitialized
}


def check_status(status, *args, **kwargs):
    """
    Look up the status return code from an API call and raise the exception that matches
    """
    ex = STATUS_EXCEPTION_MAP.get(status, None)
    if ex is not None:
        raise ex(*args, **kwargs)

__all__ = ['Status']
